from collections.abc import Mapping
from pathlib import Path
from typing import Any, Optional

import click
import yaml

from dagster_components.core.component import Component
from dagster_components.core.component_scaffolder import (
    ComponentScaffolderUnavailableReason,
    ComponentScaffoldRequest,
)


class ComponentDumper(yaml.Dumper):
    def write_line_break(self) -> None:
        # add an extra line break between top-level keys
        if self.indent == 0:
            super().write_line_break()
        super().write_line_break()


def scaffold_component_yaml(
    request: ComponentScaffoldRequest, component_params: Optional[Mapping[str, Any]]
) -> None:
    with open(request.component_instance_root_path / "component.yaml", "w") as f:
        component_data = {"type": request.component_type_name, "params": component_params or {}}
        yaml.dump(
            component_data, f, Dumper=ComponentDumper, sort_keys=False, default_flow_style=False
        )
        f.writelines([""])


def scaffold_component_instance(
    path: Path,
    component_type: type[Component],
    component_type_name: str,
    scaffold_params: Mapping[str, Any],
) -> None:
    click.echo(f"Creating a Dagster component instance folder at {path}.")
    if not path.exists():
        path.mkdir()
    scaffolder = component_type.get_scaffolder()

    if isinstance(scaffolder, ComponentScaffolderUnavailableReason):
        raise Exception(
            f"Component type {component_type_name} does not have a scaffolder. Reason: {scaffolder.message}."
        )

    scaffolder.scaffold(
        ComponentScaffoldRequest(
            component_type_name=component_type_name,
            component_instance_root_path=path,
        ),
        scaffold_params,
    )

    component_yaml_path = path / "component.yaml"
    if not component_yaml_path.exists():
        raise Exception(
            f"Currently all components require a component.yaml file. Please ensure your implementation of scaffold writes this file at {component_yaml_path}."
        )



from dataclasses import dataclass
from pathlib import Path
from typing import Callable, Optional


@dataclass
class ComponentValidationTestCase:
    """A test case for validating a component, used to
    test raw YAML error messages as well as the check CLI.
    """

    component_path: str
    component_type_filepath: Optional[Path]
    should_error: bool
    validate_error_msg: Optional[Callable[[str], None]] = None
    check_error_msg: Optional[Callable[[str], None]] = None


def msg_includes_all_of(*substrings: str) -> Callable[[str], None]:
    def _validate_error_msg(msg: str) -> None:
        for substring in substrings:
            assert substring in msg, f"Expected '{substring}' to be in error message '{msg}'"

    return _validate_error_msg


BASIC_INVALID_VALUE = ComponentValidationTestCase(
    component_path="validation/basic_component_invalid_value",
    component_type_filepath=Path(__file__).parent / "basic_components.py",
    should_error=True,
    validate_error_msg=msg_includes_all_of(
        "component.yaml:5", "params.an_int", "Input should be a valid integer"
    ),
    check_error_msg=msg_includes_all_of(
        "component.yaml:5",
        "params.an_int",
        "{} is not of type 'integer'",
    ),
)

BASIC_MISSING_VALUE = ComponentValidationTestCase(
    component_path="validation/basic_component_missing_value",
    component_type_filepath=Path(__file__).parent / "basic_components.py",
    should_error=True,
    validate_error_msg=msg_includes_all_of("component.yaml:3", "params.an_int", "required"),
    check_error_msg=msg_includes_all_of(
        "component.yaml:3",
        "params",
        "'an_int' is a required property",
    ),
)

BASIC_VALID_VALUE = ComponentValidationTestCase(
    component_path="validation/basic_component_success",
    component_type_filepath=Path(__file__).parent / "basic_components.py",
    should_error=False,
)

COMPONENT_VALIDATION_TEST_CASES = [
    BASIC_VALID_VALUE,
    BASIC_INVALID_VALUE,
    BASIC_MISSING_VALUE,
    ComponentValidationTestCase(
        component_path="validation/simple_asset_invalid_value",
        component_type_filepath=None,
        should_error=True,
        validate_error_msg=msg_includes_all_of(
            "component.yaml:5", "params.value", "Input should be a valid string"
        ),
        check_error_msg=msg_includes_all_of(
            "component.yaml:5",
            "params.value",
            "{} is not of type 'string'",
        ),
    ),
    ComponentValidationTestCase(
        component_path="validation/basic_component_extra_value",
        component_type_filepath=Path(__file__).parent / "basic_components.py",
        should_error=True,
        validate_error_msg=msg_includes_all_of(
            "component.yaml:7", "params.a_bool", "Extra inputs are not permitted"
        ),
        check_error_msg=msg_includes_all_of(
            "component.yaml:3",
            "'a_bool' was unexpected",
        ),
    ),
    ComponentValidationTestCase(
        component_path="validation/nested_component_invalid_values",
        component_type_filepath=Path(__file__).parent / "basic_components.py",
        should_error=True,
        validate_error_msg=msg_includes_all_of(
            "component.yaml:7",
            "params.nested.foo.an_int",
            "Input should be a valid integer",
            "component.yaml:12",
            "params.nested.baz.a_string",
            "Input should be a valid string",
        ),
        check_error_msg=msg_includes_all_of(
            "component.yaml:7",
            "params.nested.foo.an_int",
            "{} is not of type 'integer'",
            "component.yaml:12",
            "params.nested.baz.a_string",
            "{} is not of type 'string'",
        ),
    ),
    ComponentValidationTestCase(
        component_path="validation/nested_component_missing_values",
        component_type_filepath=Path(__file__).parent / "basic_components.py",
        should_error=True,
        validate_error_msg=msg_includes_all_of(
            "component.yaml:5", "params.nested.foo.an_int", "required"
        ),
        check_error_msg=msg_includes_all_of(
            "component.yaml:5",
            "params.nested.foo",
            "'an_int' is a required property",
            "component.yaml:10",
            "params.nested.baz",
            "'a_string' is a required property",
        ),
    ),
    ComponentValidationTestCase(
        component_path="validation/nested_component_extra_values",
        component_type_filepath=Path(__file__).parent / "basic_components.py",
        should_error=True,
        validate_error_msg=msg_includes_all_of(
            "component.yaml:8",
            "params.nested.foo.a_bool",
            "Extra inputs are not permitted",
            "component.yaml:15",
            "params.nested.baz.another_bool",
        ),
        check_error_msg=msg_includes_all_of(
            "component.yaml:5",
            "params.nested.foo",
            "'a_bool' was unexpected",
            "component.yaml:12",
            "params.nested.baz",
            "'another_bool' was unexpected",
        ),
    ),
    ComponentValidationTestCase(
        component_path="validation/invalid_component_file_model",
        component_type_filepath=Path(__file__).parent / "basic_components.py",
        should_error=True,
        validate_error_msg=msg_includes_all_of(
            "component.yaml:1",
            "type",
            "Input should be a valid string",
            "component.yaml:3",
            "params",
            "Input should be an object",
        ),
        check_error_msg=msg_includes_all_of(
            "component.yaml:1",
            "type",
            "{} is not of type 'string'",
            "component.yaml:3",
            "params",
            "'asdfasdf' is not of type 'object'",
        ),
    ),
    ComponentValidationTestCase(
        component_path="validation/invalid_yaml_missing_quote",
        component_type_filepath=None,
        should_error=True,
        validate_error_msg=msg_includes_all_of(
            "line 2",
            "found unexpected end of stream",
        ),
        check_error_msg=msg_includes_all_of(
            "component.yaml:2",
            "Unable to parse YAML",
            "found unexpected end of stream",
        ),
    ),
    ComponentValidationTestCase(
        component_path="validation/invalid_yaml_invalid_char",
        component_type_filepath=None,
        should_error=True,
        validate_error_msg=msg_includes_all_of(
            "line 1",
            "found character '@' that cannot start any token",
        ),
        check_error_msg=msg_includes_all_of(
            "component.yaml:1",
            "Unable to parse YAML",
            "found character '@' that cannot start any token",
        ),
    ),
]



from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.decorators.asset_decorator import asset
from dagster._core.definitions.definitions_class import Definitions
from dagster._core.execution.context.asset_execution_context import AssetExecutionContext
from pydantic import BaseModel
from typing_extensions import Self

from dagster_components import Component, ComponentLoadContext, registered_component_type
from dagster_components.core.component_scaffolder import (
    ComponentScaffolder,
    DefaultComponentScaffolder,
)


class SimpleAssetParams(BaseModel):
    asset_key: str
    value: str


@registered_component_type(name="simple_asset")
class SimpleAsset(Component):
    """A simple asset that returns a constant string value."""

    @classmethod
    def get_schema(cls):
        return SimpleAssetParams

    @classmethod
    def get_scaffolder(cls) -> ComponentScaffolder:
        return DefaultComponentScaffolder()

    @classmethod
    def load(cls, params: SimpleAssetParams, context: "ComponentLoadContext") -> Self:
        return cls(
            asset_key=AssetKey.from_user_string(params.asset_key),
            value=params.value,
        )

    def __init__(self, asset_key: AssetKey, value: str):
        self._asset_key = asset_key
        self._value = value

    def build_defs(self, context: ComponentLoadContext) -> Definitions:
        @asset(key=self._asset_key)
        def dummy(context: AssetExecutionContext):
            return self._value

        return Definitions(assets=[dummy])



from dagster_components.core.component import (
    Component as Component,
    ComponentLoadContext as ComponentLoadContext,
    ComponentTypeRegistry as ComponentTypeRegistry,
    registered_component_type as registered_component_type,
)
from dagster_components.core.component_defs_builder import (
    build_component_defs as build_component_defs,
)
from dagster_components.core.component_scaffolder import (
    ComponentScaffolder as ComponentScaffolder,
    ComponentScaffolderUnavailableReason as ComponentScaffolderUnavailableReason,
    ComponentScaffoldRequest as ComponentScaffoldRequest,
    DefaultComponentScaffolder as DefaultComponentScaffolder,
)
from dagster_components.core.schema.base import ResolvableModel as ResolvableModel
from dagster_components.core.schema.metadata import ResolvableFieldInfo as ResolvableFieldInfo
from dagster_components.core.schema.objects import (
    AssetAttributesModel as AssetAttributesModel,
    AssetSpecTransformModel as AssetSpecTransformModel,
    OpSpecModel as OpSpecModel,
)
from dagster_components.core.schema.resolver import TemplatedValueResolver as TemplatedValueResolver
from dagster_components.version import __version__ as __version__



import copy
import dataclasses
import importlib
import importlib.metadata
import inspect
import sys
import textwrap
from abc import ABC, abstractmethod
from collections.abc import Iterable, Mapping, Sequence
from dataclasses import dataclass
from pathlib import Path
from types import ModuleType
from typing import Any, Callable, ClassVar, Optional, TypedDict, TypeVar, Union

from dagster import _check as check
from dagster._core.definitions.definitions_class import Definitions
from dagster._core.errors import DagsterError
from dagster._utils import snakecase
from pydantic import BaseModel
from typing_extensions import Self

from dagster_components.core.component_key import (
    ComponentKey,
    GlobalComponentKey,
    LocalComponentKey,
)
from dagster_components.core.component_scaffolder import (
    ComponentScaffolder,
    ComponentScaffolderUnavailableReason,
    DefaultComponentScaffolder,
)
from dagster_components.core.schema.resolver import TemplatedValueResolver
from dagster_components.utils import load_module_from_path


class ComponentDeclNode(ABC):
    @abstractmethod
    def load(self, context: "ComponentLoadContext") -> Sequence["Component"]: ...


class Component(ABC):
    name: ClassVar[Optional[str]] = None

    @classmethod
    def get_schema(cls) -> Optional[type[BaseModel]]:
        return None

    @classmethod
    def get_scaffolder(cls) -> Union[ComponentScaffolder, ComponentScaffolderUnavailableReason]:
        """Subclasses should implement this method to override scaffolding behavior. If this component
        is not meant to be scaffolded it returns a ComponentScaffolderUnavailableReason with a message
        This can be determined at runtime based on the environment or configuration. For example,
        if scaffolders are optionally installed as extras (for example to avoid heavy dependencies in production),
        this method should return a ComponentScaffolderUnavailableReason with a message explaining
        how to install the necessary extras.
        """
        return DefaultComponentScaffolder()

    @classmethod
    def get_additional_scope(cls) -> Mapping[str, Any]:
        return {}

    @abstractmethod
    def build_defs(self, context: "ComponentLoadContext") -> Definitions: ...

    @classmethod
    @abstractmethod
    def load(cls, params: Optional[BaseModel], context: "ComponentLoadContext") -> Self: ...

    @classmethod
    def get_metadata(cls) -> "ComponentTypeInternalMetadata":
        docstring = cls.__doc__
        clean_docstring = _clean_docstring(docstring) if docstring else None

        scaffolder = cls.get_scaffolder()

        if isinstance(scaffolder, ComponentScaffolderUnavailableReason):
            raise DagsterError(
                f"Component {cls.__name__} is not scaffoldable: {scaffolder.message}"
            )

        component_params = cls.get_schema()
        scaffold_params = scaffolder.get_schema()
        return {
            "summary": clean_docstring.split("\n\n")[0] if clean_docstring else None,
            "description": clean_docstring if clean_docstring else None,
            "scaffold_params_schema": None
            if scaffold_params is None
            else scaffold_params.model_json_schema(),
            "component_params_schema": None
            if component_params is None
            else component_params.model_json_schema(),
        }

    @classmethod
    def get_description(cls) -> Optional[str]:
        return inspect.getdoc(cls)


def _clean_docstring(docstring: str) -> str:
    lines = docstring.strip().splitlines()
    first_line = lines[0]
    if len(lines) == 1:
        return first_line
    else:
        rest = textwrap.dedent("\n".join(lines[1:]))
        return f"{first_line}\n{rest}"


class ComponentTypeInternalMetadata(TypedDict):
    summary: Optional[str]
    description: Optional[str]
    scaffold_params_schema: Optional[Any]  # json schema
    component_params_schema: Optional[Any]  # json schema


class ComponentTypeMetadata(ComponentTypeInternalMetadata):
    name: str
    namespace: str


def get_entry_points_from_python_environment(group: str) -> Sequence[importlib.metadata.EntryPoint]:
    if sys.version_info >= (3, 10):
        return importlib.metadata.entry_points(group=group)
    else:
        return importlib.metadata.entry_points().get(group, [])


COMPONENTS_ENTRY_POINT_GROUP = "dagster.components"
BUILTIN_COMPONENTS_ENTRY_POINT_BASE = "dagster_components"
BUILTIN_MAIN_COMPONENT_ENTRY_POINT = BUILTIN_COMPONENTS_ENTRY_POINT_BASE
BUILTIN_TEST_COMPONENT_ENTRY_POINT = ".".join([BUILTIN_COMPONENTS_ENTRY_POINT_BASE, "test"])


class ComponentTypeRegistry:
    @classmethod
    def from_entry_point_discovery(
        cls, builtin_component_lib: str = BUILTIN_MAIN_COMPONENT_ENTRY_POINT
    ) -> "ComponentTypeRegistry":
        """Discover component types registered in the Python environment via the
        `dagster_components` entry point group.

        `dagster-components` itself registers multiple component entry points. We call these
        "builtin" component libraries. The `dagster_components` entry point resolves to published
        component types and is loaded by default. Other entry points resolve to various sets of test
        component types. This method will only ever load one builtin component library.

        Args:
            builtin-component-lib (str): Specifies the builtin components library to load. Built-in
            component libraries are defined under entry points with names matching the pattern
            `dagster_components*`. Only one built-in  component library can be loaded at a time.
            Defaults to `dagster_components`, the standard set of published component types.
        """
        component_types: dict[ComponentKey, type[Component]] = {}
        for entry_point in get_entry_points_from_python_environment(COMPONENTS_ENTRY_POINT_GROUP):
            # Skip built-in entry points that are not the specified builtin component library.
            if (
                entry_point.name.startswith(BUILTIN_COMPONENTS_ENTRY_POINT_BASE)
                and not entry_point.name == builtin_component_lib
            ):
                continue

            root_module = entry_point.load()
            if not isinstance(root_module, ModuleType):
                raise DagsterError(
                    f"Invalid entry point {entry_point.name} in group {COMPONENTS_ENTRY_POINT_GROUP}. "
                    f"Value expected to be a module, got {root_module}."
                )
            for component_type in get_registered_component_types_in_module(root_module):
                key = GlobalComponentKey(
                    name=get_component_type_name(component_type), namespace=entry_point.name
                )
                component_types[key] = component_type

        return cls(component_types)

    def __init__(self, component_types: dict[ComponentKey, type[Component]]):
        self._component_types: dict[ComponentKey, type[Component]] = copy.copy(component_types)

    @staticmethod
    def empty() -> "ComponentTypeRegistry":
        return ComponentTypeRegistry({})

    def register(self, key: ComponentKey, component_type: type[Component]) -> None:
        if key in self._component_types:
            raise DagsterError(f"There is an existing component registered under {key}")
        self._component_types[key] = component_type

    def has(self, key: ComponentKey) -> bool:
        return key in self._component_types

    def get(self, key: ComponentKey) -> type[Component]:
        if isinstance(key, LocalComponentKey):
            for component_type in find_component_types_in_file(key.python_file):
                if get_component_type_name(component_type) == key.name:
                    return component_type
            raise ValueError(
                f"Could not find component type {key.to_typename()} in {key.python_file}"
            )
        else:
            return self._component_types[key]

    def keys(self) -> Iterable[ComponentKey]:
        return self._component_types.keys()

    def items(self) -> Iterable[tuple[ComponentKey, type[Component]]]:
        return self._component_types.items()

    def __repr__(self) -> str:
        return f"<ComponentRegistry {list(self._component_types.keys())}>"


def get_registered_component_types_in_module(module: ModuleType) -> Iterable[type[Component]]:
    from dagster._core.definitions.module_loaders.load_assets_from_modules import (
        find_subclasses_in_module,
    )

    for component in find_subclasses_in_module(module, (Component,)):
        if is_registered_component_type(component):
            yield component


T = TypeVar("T", bound=BaseModel)


@dataclass
class ComponentLoadContext:
    resources: Mapping[str, object]
    registry: ComponentTypeRegistry
    decl_node: Optional[ComponentDeclNode]
    templated_value_resolver: TemplatedValueResolver

    @staticmethod
    def for_test(
        *,
        resources: Optional[Mapping[str, object]] = None,
        registry: Optional[ComponentTypeRegistry] = None,
        decl_node: Optional[ComponentDeclNode] = None,
    ) -> "ComponentLoadContext":
        return ComponentLoadContext(
            resources=resources or {},
            registry=registry or ComponentTypeRegistry.empty(),
            decl_node=decl_node,
            templated_value_resolver=TemplatedValueResolver.default(),
        )

    @property
    def path(self) -> Path:
        from dagster_components.core.component_decl_builder import (
            PythonComponentDecl,
            YamlComponentDecl,
        )

        if not isinstance(self.decl_node, (YamlComponentDecl, PythonComponentDecl)):
            check.failed(f"Unsupported decl_node type {type(self.decl_node)}")

        return self.decl_node.path

    def with_rendering_scope(self, rendering_scope: Mapping[str, Any]) -> "ComponentLoadContext":
        return dataclasses.replace(
            self,
            templated_value_resolver=self.templated_value_resolver.with_scope(**rendering_scope),
        )

    def for_decl_node(self, decl_node: ComponentDeclNode) -> "ComponentLoadContext":
        return dataclasses.replace(self, decl_node=decl_node)


COMPONENT_REGISTRY_KEY_ATTR = "__dagster_component_registry_key"
COMPONENT_LOADER_FN_ATTR = "__dagster_component_loader_fn"


def registered_component_type(
    cls: Optional[type[Component]] = None, *, name: Optional[str] = None
) -> Any:
    """Decorator for registering a component type. You must annotate a component
    type with this decorator in order for it to be inspectable and loaded by tools.

    Args:
        cls (Optional[Type[Component]]): The target of the decorator: the component class
            to register. The class must inherit from Component.
        name (Optional[str]): The name to register the component type under. If not
            provided, the name will be the snake-cased version of the class name. The
            name is used as a key in operations like scaffolding and loading.
    """
    if cls is None:

        def wrapper(actual_cls: type[Component]) -> type[Component]:
            check.inst_param(actual_cls, "actual_cls", type)
            setattr(
                actual_cls,
                COMPONENT_REGISTRY_KEY_ATTR,
                name or snakecase(actual_cls.__name__),
            )
            return actual_cls

        return wrapper
    else:
        # called without params
        check.inst_param(cls, "cls", type)
        setattr(cls, COMPONENT_REGISTRY_KEY_ATTR, name or snakecase(cls.__name__))
        return cls


def is_registered_component_type(cls: type) -> bool:
    return hasattr(cls, COMPONENT_REGISTRY_KEY_ATTR)


def get_component_type_name(component_type: type[Component]) -> str:
    check.param_invariant(
        is_registered_component_type(component_type),
        "component_type",
        "Expected a registered component. Use @component to register a component.",
    )
    return getattr(component_type, COMPONENT_REGISTRY_KEY_ATTR)


T_Component = TypeVar("T_Component", bound=Component)


def component_loader(
    fn: Callable[[ComponentLoadContext], T],
) -> Callable[[ComponentLoadContext], T]:
    setattr(fn, COMPONENT_LOADER_FN_ATTR, True)
    return fn


def is_component_loader(obj: Any) -> bool:
    return getattr(obj, COMPONENT_LOADER_FN_ATTR, False)


def find_component_types_in_file(file_path: Path) -> list[type[Component]]:
    """Find all component types defined in a specific file."""
    component_types = []
    for _name, obj in inspect.getmembers(
        load_module_from_path(file_path.stem, file_path), inspect.isclass
    ):
        assert isinstance(obj, type)
        if is_registered_component_type(obj):
            component_types.append(obj)
    return component_types


def find_local_component_types(component_path: Path) -> Mapping[LocalComponentKey, type[Component]]:
    """Find all component types defined in a component directory, and their respective paths."""
    component_types = {}
    for py_file in component_path.glob("*.py"):
        for component_type in find_component_types_in_file(py_file):
            component_types[
                LocalComponentKey(
                    name=get_component_type_name(component_type),
                    namespace=py_file.name,
                    dirpath=py_file.parent,
                )
            ] = component_type
    return component_types



from abc import abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Optional

from dagster._record import record
from pydantic import BaseModel


@record
class ComponentScaffoldRequest:
    component_type_name: str
    component_instance_root_path: Path


class ComponentScaffolder:
    @classmethod
    def get_schema(cls) -> Optional[type[BaseModel]]:
        return None

    @abstractmethod
    def scaffold(self, request: ComponentScaffoldRequest, params: Any) -> None: ...


@dataclass
class ComponentScaffolderUnavailableReason:
    message: str


class DefaultComponentScaffolder(ComponentScaffolder):
    def scaffold(self, request: ComponentScaffoldRequest, params: Any) -> None:
        # This will be deleted once all components are converted to the new ComponentScaffolder API
        from dagster_components.scaffold import scaffold_component_yaml

        scaffold_component_yaml(request, params.model_dump() if params else {})



from collections.abc import Iterator, Sequence
from typing import Annotated, Callable, Optional

from dagster._core.definitions.definitions_class import Definitions
from dagster._core.execution.context.asset_execution_context import AssetExecutionContext
from dagster_dbt import (
    DagsterDbtTranslator,
    DbtCliResource,
    DbtManifestAssetSelection,
    DbtProject,
    dbt_assets,
)

from dagster_components import Component, ComponentLoadContext
from dagster_components.core.component import registered_component_type
from dagster_components.core.schema.base import ResolvableModel
from dagster_components.core.schema.metadata import ResolvableFieldInfo
from dagster_components.core.schema.objects import (
    AssetAttributesModel,
    AssetSpecTransformModel,
    OpSpecModel,
    TemplatedValueResolver,
)
from dagster_components.lib.dbt_project.scaffolder import DbtProjectComponentScaffolder
from dagster_components.utils import ResolvingInfo, get_wrapped_translator_class


class DbtProjectParams(ResolvableModel["DbtProjectComponent"]):
    dbt: DbtCliResource
    op: Optional[OpSpecModel] = None
    asset_attributes: Annotated[
        Optional[AssetAttributesModel], ResolvableFieldInfo(required_scope={"node"})
    ] = None
    transforms: Optional[Sequence[AssetSpecTransformModel]] = None

    def resolve(self, resolver: TemplatedValueResolver) -> "DbtProjectComponent":
        return DbtProjectComponent(
            resource=self.dbt,
            op_spec=self.op.resolve(resolver) if self.op else None,
            translator=get_wrapped_translator_class(DagsterDbtTranslator)(
                resolving_info=ResolvingInfo(
                    "node", self.asset_attributes or AssetAttributesModel(), resolver
                )
            ),
            transforms=[transform.resolve(resolver) for transform in (self.transforms or [])],
        )


@registered_component_type(name="dbt_project")
class DbtProjectComponent(Component):
    """Expose a DBT project to Dagster as a set of assets."""

    def __init__(
        self,
        resource: DbtCliResource,
        op_spec: Optional[OpSpecModel],
        translator: DagsterDbtTranslator,
        transforms: Sequence[Callable[[Definitions], Definitions]],
    ):
        self.resource = resource
        self.project = DbtProject(resource.project_dir)
        self.op_spec = op_spec
        self.transforms = transforms
        self.translator = translator

    @classmethod
    def get_scaffolder(cls) -> "DbtProjectComponentScaffolder":
        return DbtProjectComponentScaffolder()

    @classmethod
    def get_schema(cls) -> type[DbtProjectParams]:
        return DbtProjectParams

    @classmethod
    def load(cls, params: DbtProjectParams, context: ComponentLoadContext) -> "DbtProjectComponent":
        return params.resolve(context.templated_value_resolver)

    def get_asset_selection(
        self, select: str, exclude: Optional[str] = None
    ) -> DbtManifestAssetSelection:
        return DbtManifestAssetSelection.build(
            manifest=self.project.manifest_path,
            dagster_dbt_translator=self.translator,
            select=select,
            exclude=exclude,
        )

    def build_defs(self, context: ComponentLoadContext) -> Definitions:
        self.project.prepare_if_dev()

        @dbt_assets(
            manifest=self.project.manifest_path,
            project=self.project,
            name=self.op_spec.name if self.op_spec else self.project.name,
            op_tags=self.op_spec.tags if self.op_spec else None,
            dagster_dbt_translator=self.translator,
        )
        def _fn(context: AssetExecutionContext):
            yield from self.execute(context=context, dbt=self.resource)

        defs = Definitions(assets=[_fn])
        for transform in self.transforms:
            defs = transform(defs)
        return defs

    def execute(self, context: AssetExecutionContext, dbt: DbtCliResource) -> Iterator:
        yield from dbt.cli(["build"], context=context).stream()



from pathlib import Path
from typing import Optional

import click
from pydantic import TypeAdapter

from dagster_components import ComponentTypeRegistry
from dagster_components.core.component_key import GlobalComponentKey
from dagster_components.scaffold import (
    ComponentScaffolderUnavailableReason,
    scaffold_component_instance,
)
from dagster_components.utils import CLI_BUILTIN_COMPONENT_LIB_KEY, exit_with_error


@click.group(name="scaffold")
def scaffold_cli() -> None:
    """Commands for scaffolding Dagster components and related entities."""


@scaffold_cli.command(name="component")
@click.argument("component_type", type=str)
@click.argument("component_path", type=Path)
@click.option("--json-params", type=str, default=None)
@click.pass_context
def scaffold_component_command(
    ctx: click.Context,
    component_type: str,
    component_path: Path,
    json_params: Optional[str],
) -> None:
    builtin_component_lib = ctx.obj.get(CLI_BUILTIN_COMPONENT_LIB_KEY, False)
    registry = ComponentTypeRegistry.from_entry_point_discovery(
        builtin_component_lib=builtin_component_lib
    )
    component_key = GlobalComponentKey.from_typename(component_type)
    if not registry.has(component_key):
        exit_with_error(f"No component type `{component_type}` could be resolved.")

    component_type_cls = registry.get(component_key)
    if json_params:
        scaffolder = component_type_cls.get_scaffolder()
        if isinstance(scaffolder, ComponentScaffolderUnavailableReason):
            raise Exception(
                f"Component type {component_type} does not have a scaffolder. Reason: {scaffolder.message}."
            )
        scaffold_params = TypeAdapter(scaffolder.get_schema()).validate_json(json_params)
    else:
        scaffold_params = {}

    scaffold_component_instance(
        component_path,
        component_type_cls,
        component_type,
        scaffold_params,
    )



import os
from pathlib import Path
from typing import Any

import yaml

from dagster_components.core.component_scaffolder import (
    ComponentScaffolder,
    ComponentScaffoldRequest,
)
from dagster_components.scaffold import scaffold_component_yaml


class SlingReplicationComponentScaffolder(ComponentScaffolder):
    def scaffold(self, request: ComponentScaffoldRequest, params: Any) -> None:
        scaffold_component_yaml(request, {"replications": [{"path": "replication.yaml"}]})
        replication_path = Path(os.getcwd()) / "replication.yaml"
        with open(replication_path, "w") as f:
            yaml.dump({"source": {}, "target": {}, "streams": {}}, f)



import shutil
from pathlib import Path

from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.decorators.asset_decorator import asset
from dagster._core.definitions.definitions_class import Definitions
from dagster._core.execution.context.asset_execution_context import AssetExecutionContext
from dagster._core.pipes.subprocess import PipesSubprocessClient
from pydantic import BaseModel
from typing_extensions import Self

from dagster_components import Component, ComponentLoadContext, registered_component_type
from dagster_components.core.component_scaffolder import (
    ComponentScaffolder,
    ComponentScaffoldRequest,
)
from dagster_components.scaffold import scaffold_component_yaml


# Same schema used for file generation and defs generation
class SimplePipesScriptAssetParams(BaseModel):
    asset_key: str
    filename: str


class SimplePipesScriptAssetScaffolder(ComponentScaffolder):
    @classmethod
    def get_schema(cls):
        return SimplePipesScriptAssetParams

    def scaffold(
        self, request: ComponentScaffoldRequest, params: SimplePipesScriptAssetParams
    ) -> None:
        scaffold_component_yaml(request, params.model_dump())
        Path(request.component_instance_root_path, params.filename).write_text(
            _SCRIPT_TEMPLATE.format(asset_key=params.asset_key)
        )


_SCRIPT_TEMPLATE = """
from dagster_pipes import open_dagster_pipes

context = open_dagster_pipes()

context.log.info("Materializing asset {asset_key} from pipes")
context.report_asset_materialization(asset_key="{asset_key}")
"""


@registered_component_type(name="simple_pipes_script_asset")
class SimplePipesScriptAsset(Component):
    """A simple asset that runs a Python script with the Pipes subprocess client.

    Because it is a pipes asset, no value is returned.
    """

    @classmethod
    def get_scaffolder(cls) -> ComponentScaffolder:
        return SimplePipesScriptAssetScaffolder()

    @classmethod
    def get_schema(cls):
        return SimplePipesScriptAssetParams

    @classmethod
    def load(cls, params: SimplePipesScriptAssetParams, context: "ComponentLoadContext") -> Self:
        return cls(
            asset_key=AssetKey.from_user_string(params.asset_key),
            script_path=context.path / params.filename,
        )

    def __init__(self, asset_key: AssetKey, script_path: Path):
        self._asset_key = asset_key
        self._script_path = script_path

    def build_defs(self, context: ComponentLoadContext) -> Definitions:
        @asset(key=self._asset_key)
        def _asset(context: AssetExecutionContext, pipes_client: PipesSubprocessClient):
            cmd = [shutil.which("python"), self._script_path]
            return pipes_client.run(command=cmd, context=context).get_results()

        return Definitions(assets=[_asset])



import json
from collections.abc import Sequence
from pathlib import Path
from typing import Any, Literal, Union

import click
from pydantic import TypeAdapter, create_model

from dagster_components.core.component import (
    ComponentTypeMetadata,
    ComponentTypeRegistry,
    find_local_component_types,
)
from dagster_components.utils import CLI_BUILTIN_COMPONENT_LIB_KEY


@click.group(name="list")
def list_cli():
    """Commands for listing Dagster components and related entities."""


@list_cli.command(name="component-types")
@click.pass_context
def list_component_types_command(ctx: click.Context) -> None:
    """List registered Dagster components."""
    builtin_component_lib = ctx.obj.get(CLI_BUILTIN_COMPONENT_LIB_KEY, False)
    output: dict[str, Any] = {}
    registry = ComponentTypeRegistry.from_entry_point_discovery(
        builtin_component_lib=builtin_component_lib
    )
    for key in sorted(registry.keys(), key=lambda k: k.to_typename()):
        output[key.to_typename()] = ComponentTypeMetadata(
            name=key.name,
            namespace=key.namespace,
            **registry.get(key).get_metadata(),
        )
    click.echo(json.dumps(output))


@list_cli.command(name="local-component-types")
@click.argument("component_directories", nargs=-1, type=click.Path(exists=True))
def list_local_component_types_command(component_directories: Sequence[str]) -> None:
    """List local Dagster components found in the specified directories."""
    output: dict = {}
    for component_directory in component_directories:
        output_for_directory = {}
        for key, component_type in find_local_component_types(Path(component_directory)).items():
            output_for_directory[key.to_typename()] = ComponentTypeMetadata(
                name=key.name,
                namespace=key.namespace,
                **component_type.get_metadata(),
            )
        if len(output_for_directory) > 0:
            output[component_directory] = output_for_directory
    click.echo(json.dumps(output))


@list_cli.command(name="all-components-schema")
@click.pass_context
def list_all_components_schema_command(ctx: click.Context) -> None:
    """Builds a JSON schema which ORs the schema for a component
    file for all component types available in the current code location.
    """
    builtin_component_lib = ctx.obj.get(CLI_BUILTIN_COMPONENT_LIB_KEY, False)
    registry = ComponentTypeRegistry.from_entry_point_discovery(
        builtin_component_lib=builtin_component_lib
    )

    schemas = []
    for key in sorted(registry.keys(), key=lambda k: k.to_typename()):
        component_type = registry.get(key)
        # Create ComponentFileModel schema for each type
        schema_type = component_type.get_schema()
        key_string = key.to_typename()
        if schema_type:
            schemas.append(
                create_model(
                    key.name, type=(Literal[key_string], key_string), params=(schema_type, None)
                )
            )
    union_type = Union[tuple(schemas)]  # type: ignore
    click.echo(json.dumps(TypeAdapter(union_type).json_schema()))



from pathlib import Path
from typing import Optional

from dagster._utils import pushd
from pydantic import BaseModel

from dagster_components.core.component import ComponentScaffolder
from dagster_components.core.component_scaffolder import ComponentScaffoldRequest
from dagster_components.scaffold import scaffold_component_yaml


class DefinitionsScaffoldParams(BaseModel):
    definitions_path: Optional[str] = None


class DefinitionsComponentScaffolder(ComponentScaffolder):
    @classmethod
    def get_schema(cls):
        return DefinitionsScaffoldParams

    def scaffold(
        self, request: ComponentScaffoldRequest, params: DefinitionsScaffoldParams
    ) -> None:
        scaffold_params = (
            params if isinstance(params, DefinitionsScaffoldParams) else DefinitionsScaffoldParams()
        )

        with pushd(str(request.component_instance_root_path)):
            Path(
                scaffold_params.definitions_path
                if scaffold_params.definitions_path
                else "definitions.py"
            ).touch(exist_ok=True)

        scaffold_component_yaml(
            request,
            {"definitions_path": scaffold_params.definitions_path}
            if scaffold_params.definitions_path
            else {},
        )



from dagster._core.definitions.decorators.asset_decorator import asset
from dagster._core.definitions.definitions_class import Definitions
from dagster._core.execution.context.asset_execution_context import AssetExecutionContext
from typing_extensions import Self

from dagster_components import Component, ComponentLoadContext, registered_component_type
from dagster_components.core.component_scaffolder import DefaultComponentScaffolder


@registered_component_type(name="all_metadata_empty_asset")
class AllMetadataEmptyAsset(Component):
    @classmethod
    def load(cls, context: "ComponentLoadContext") -> Self:
        return cls()

    @classmethod
    def get_scaffolder(cls) -> DefaultComponentScaffolder:
        return DefaultComponentScaffolder()

    def build_defs(self, context: ComponentLoadContext) -> Definitions:
        @asset
        def hardcoded_asset(context: AssetExecutionContext):
            return 1

        return Definitions(assets=[hardcoded_asset])



from collections.abc import Iterator, Mapping, Sequence, Set
from dataclasses import dataclass
from typing import Annotated, Any, Callable, Optional, Union, get_args, get_origin

import dagster._check as check
from pydantic.fields import FieldInfo

REF_BASE = "#/$defs/"
REF_TEMPLATE = f"{REF_BASE}{{model}}"
JSON_SCHEMA_EXTRA_REQUIRED_SCOPE_KEY = "dagster_required_scope"


@dataclass
class ResolutionMetadata:
    """Internal class that stores arbitrary metadata about a resolved field."""

    output_type: type
    post_process: Optional[Callable[[Any], Any]] = None


class ResolvableFieldInfo(FieldInfo):
    """Wrapper class that stores additional resolution metadata within a pydantic FieldInfo object.

    Examples:
    ```python
    class MyModel(ResolvableModel):
        resolvable_obj: Annotated[str, ResolvableFieldInfo(output_type=SomeObj)]
    ```
    """

    def __init__(
        self,
        *,
        output_type: Optional[type] = None,
        post_process_fn: Optional[Callable[[Any], Any]] = None,
        required_scope: Optional[Set[str]] = None,
    ):
        self.resolution_metadata = (
            ResolutionMetadata(output_type=output_type, post_process=post_process_fn)
            if output_type
            else None
        )
        super().__init__(
            json_schema_extra={JSON_SCHEMA_EXTRA_REQUIRED_SCOPE_KEY: list(required_scope or [])},
        )


def get_resolution_metadata(annotation: type) -> ResolutionMetadata:
    origin = get_origin(annotation)
    if origin is Annotated:
        _, f_metadata, *_ = get_args(annotation)
        if isinstance(f_metadata, ResolvableFieldInfo) and f_metadata.resolution_metadata:
            return f_metadata.resolution_metadata
    return ResolutionMetadata(output_type=annotation)


def _subschemas_on_path(
    valpath: Sequence[Union[str, int]], json_schema: Mapping[str, Any], subschema: Mapping[str, Any]
) -> Iterator[Mapping[str, Any]]:
    """Given a valpath and the json schema of a given target type, returns the subschemas at each step of the path."""
    # List[ComplexType] (e.g.) will contain a reference to the complex type schema in the
    # top-level $defs, so we dereference it here.
    if "$ref" in subschema:
        # depending on the pydantic version, the extras may be stored with the reference or not
        extras = {k: v for k, v in subschema.items() if k != "$ref"}
        subschema = {**json_schema["$defs"].get(subschema["$ref"][len(REF_BASE) :]), **extras}

    yield subschema
    if len(valpath) == 0:
        return

    # Optional[ComplexType] (e.g.) will contain multiple schemas in the "anyOf" field
    if "anyOf" in subschema:
        for inner in subschema["anyOf"]:
            yield from _subschemas_on_path(valpath, json_schema, inner)

    el = valpath[0]
    if isinstance(el, str):
        # valpath: ['field']
        # field: X
        inner = subschema.get("properties", {}).get(el)
    elif isinstance(el, int):
        # valpath: ['field', 0]
        # field: List[X]
        inner = subschema.get("items")
    else:
        check.failed(f"Unexpected valpath element: {el}")

    # the path wasn't valid, or unspecified
    if not inner:
        return

    _, *rest = valpath
    yield from _subschemas_on_path(rest, json_schema, inner)


def _get_additional_required_scope(subschema: Mapping[str, Any]) -> Set[str]:
    raw = check.opt_inst(subschema.get(JSON_SCHEMA_EXTRA_REQUIRED_SCOPE_KEY), list)
    return set(raw) if raw else set()


def get_required_scope(
    valpath: Sequence[Union[str, int]], json_schema: Mapping[str, Any]
) -> Set[str]:
    """Given a valpath and the json schema of a given target type, determines the available rendering scope."""
    required_scope = set()
    for subschema in _subschemas_on_path(valpath, json_schema, json_schema):
        required_scope |= _get_additional_required_scope(subschema)
    return required_scope



import re
from abc import ABC
from dataclasses import dataclass
from pathlib import Path

COMPONENT_TYPE_REGEX = re.compile(r"^([a-zA-Z0-9_]+)@([a-zA-Z0-9_\.]+)$")


def _name_and_namespace_from_type(typename: str) -> tuple[str, str]:
    match = COMPONENT_TYPE_REGEX.match(typename)
    if not match:
        raise ValueError(f"Invalid component type name: {typename}")
    return match.group(1), match.group(2)


@dataclass(frozen=True)
class ComponentKey(ABC):
    name: str
    namespace: str

    def to_typename(self) -> str:
        return f"{self.name}@{self.namespace}"

    @staticmethod
    def from_typename(typename: str, dirpath: Path) -> "ComponentKey":
        if typename.endswith(".py"):
            return LocalComponentKey.from_type(typename, dirpath)
        else:
            return GlobalComponentKey.from_typename(typename)


@dataclass(frozen=True)
class GlobalComponentKey(ComponentKey):
    @staticmethod
    def from_typename(typename: str) -> "GlobalComponentKey":
        name, namespace = _name_and_namespace_from_type(typename)
        return GlobalComponentKey(name=name, namespace=namespace)


@dataclass(frozen=True)
class LocalComponentKey(ComponentKey):
    dirpath: Path

    @staticmethod
    def from_type(typename: str, dirpath: Path) -> "LocalComponentKey":
        name, namespace = _name_and_namespace_from_type(typename)
        return LocalComponentKey(name=name, namespace=namespace, dirpath=dirpath)

    @property
    def python_file(self) -> Path:
        return self.dirpath / self.namespace



from pathlib import Path

from setuptools import find_packages, setup


def get_version() -> str:
    version: dict[str, str] = {}
    with open(Path(__file__).parent / "dagster_components/version.py", encoding="utf8") as fp:
        exec(fp.read(), version)

    return version["__version__"]


ver = get_version()
# dont pin dev installs to avoid pip dep renderer issues
pin = "" if ver == "1!0+dev" else f"=={ver}"
setup(
    name="dagster-components",
    version=get_version(),
    author="Dagster Labs",
    author_email="hello@dagsterlabs.com",
    license="Apache-2.0",
    description="",  # TODO - fill out description
    url=(
        "https://github.com/dagster-io/dagster/tree/master/examples/experimental/dagster-components/"
        "dagster-components"
    ),
    classifiers=[
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(exclude=["dagster_components_tests*", "examples*"]),
    install_requires=["dagster>=1.9.5", "tomli", "typer"],
    zip_safe=False,
    entry_points={
        "console_scripts": [
            "dagster-components = dagster_components.cli:main",
        ],
        "dagster.components": [
            "dagster_components = dagster_components.lib",
            "dagster_components.test = dagster_components.lib.test",
        ],
    },
    extras_require={
        "sling": ["dagster-sling"],
        "dbt": ["dagster-dbt"],
        "test": ["dbt-duckdb", "dagster-dg"],
    },
)



import os
from collections.abc import Mapping, Sequence
from typing import Any, Callable, Optional, TypeVar, Union

from dagster._core.definitions.declarative_automation.automation_condition import (
    AutomationCondition,
)
from dagster._record import record
from jinja2.nativetypes import NativeTemplate

T = TypeVar("T")


def env_scope(key: str) -> Optional[str]:
    return os.environ.get(key)


def automation_condition_scope() -> Mapping[str, Any]:
    return {
        "eager": AutomationCondition.eager,
        "on_cron": AutomationCondition.on_cron,
    }


@record
class TemplatedValueResolver:
    scope: Mapping[str, Any]

    @staticmethod
    def default() -> "TemplatedValueResolver":
        return TemplatedValueResolver(
            scope={"env": env_scope, "automation_condition": automation_condition_scope()}
        )

    def with_scope(self, **additional_scope) -> "TemplatedValueResolver":
        return TemplatedValueResolver(scope={**self.scope, **additional_scope})

    def _resolve_value(self, val: Any) -> Any:
        """Resolves a single value, if it is a templated string."""
        return NativeTemplate(val).render(**self.scope) if isinstance(val, str) else val

    def _resolve_obj(
        self,
        obj: Any,
        valpath: Optional[Sequence[Union[str, int]]],
        should_resolve: Callable[[Sequence[Union[str, int]]], bool],
    ) -> Any:
        """Recursively resolves templated values in a nested object, based on the provided should_resolve function."""
        if valpath is not None and not should_resolve(valpath):
            return obj
        elif isinstance(obj, dict):
            # resolve all values in the dict
            return {
                k: self._resolve_obj(
                    v, [*valpath, k] if valpath is not None else None, should_resolve
                )
                for k, v in obj.items()
            }
        elif isinstance(obj, list):
            # resolve all values in the list
            return [
                self._resolve_obj(v, [*valpath, i] if valpath is not None else None, should_resolve)
                for i, v in enumerate(obj)
            ]
        else:
            return self._resolve_value(obj)

    def resolve_obj(self, val: Any) -> Any:
        """Recursively resolves templated values in a nested object."""
        return self._resolve_obj(val, None, lambda _: True)



from collections.abc import Sequence
from typing import Annotated, Optional

from dagster._core.definitions.decorators.asset_decorator import asset
from dagster._core.definitions.definitions_class import Definitions
from dagster._core.execution.context.asset_execution_context import AssetExecutionContext
from pydantic import BaseModel
from typing_extensions import Self

from dagster_components import Component, ComponentLoadContext, registered_component_type
from dagster_components.core.component_scaffolder import DefaultComponentScaffolder
from dagster_components.core.schema.metadata import ResolvableFieldInfo
from dagster_components.core.schema.objects import (
    AssetAttributesModel,
    AssetSpecTransformModel,
    OpSpecModel,
)


class ComplexAssetParams(BaseModel):
    value: str
    op: Optional[OpSpecModel] = None
    asset_attributes: Annotated[
        Optional[AssetAttributesModel], ResolvableFieldInfo(required_scope={"node"})
    ] = None
    asset_transforms: Optional[Sequence[AssetSpecTransformModel]] = None


@registered_component_type(name="complex_schema_asset")
class ComplexSchemaAsset(Component):
    """An asset that has a complex params schema."""

    @classmethod
    def get_schema(cls):
        return ComplexAssetParams

    @classmethod
    def get_scaffolder(cls) -> DefaultComponentScaffolder:
        return DefaultComponentScaffolder()

    @classmethod
    def load(cls, params: ComplexAssetParams, context: "ComponentLoadContext") -> Self:
        return cls(
            value=params.value,
            op_spec=params.op,
            asset_attributes=params.asset_attributes,
            asset_transforms=params.asset_transforms or [],
        )

    def __init__(
        self,
        value: str,
        op_spec: Optional[OpSpecModel],
        asset_attributes: Optional[AssetAttributesModel],
        asset_transforms: Sequence[AssetSpecTransformModel],
    ):
        self._value = value
        self._op_spec = op_spec
        self._asset_attributes = asset_attributes
        self._asset_transforms = asset_transforms

    def build_defs(self, context: ComponentLoadContext) -> Definitions:
        @asset(spec=self._asset_attributes)
        def dummy(context: AssetExecutionContext):
            return self._value

        return Definitions(assets=[dummy])
